const _ = require('lodash');
const semver = require('semver');
const crypto = require('crypto');
const Types = require('./Types');
const USER_AGENT_TYPES = require('./UserAgentsTypes');
const logger = require('cf-logs').Logger('codefresh:infra:authenticated-entity');
const { getInternalRequestLogic, apiService } = require('./internalRequest');

class Entity {
    constructor(data) {
        if (!data.type) {
            throw new Error('Could not create an entity without a type filed');
        }

        this._entity = data;
    }

    toJson(options = {}) {
        if (options.partial) {
            return this._getPartialJson();
        } else {
            return this._entity;
        }
    }

    toString() {
        return JSON.stringify(this._entity);
    }

    generateHashString() {
        return crypto.createHash('sha256')
            .update(this.encodeToBase64String())
            .digest('hex');
    }

    encodeToBase64String() {
        return Buffer.from(this.toString()).toString('base64');
    }

    getType() {
        return this._entity.type;
    }

    static extractAccountData(account, userId = '', insertFeatures = false) {
        return {
            name: account.name,
            id: account._id.toString(),
            _id: account._id.toString(),
            isAdmin: _.get(account, 'admins', []).includes(userId),
            adminTeam: account.adminTeam,
            teams: account.teams || [],
            paymentPlan: account.paymentPlan,
            supportPlan: account.supportPlan,
            ...(account.type && { type: account.type }),
            ...(account.trial && { trial: account.trial }),
            ...(account.segment && { segment: account.segment }),
            ...(account.pipelineConfig && { pipelineConfig: account.pipelineConfig }),
            ...(insertFeatures && account.features && { features: account.features }),
            CFCRState: {
                enabled: _.get(account, 'CFCRState.enabled', false)
            },
            sharedConfigRepo: _.get(account, 'csdp.sharedConfigRepo', ''),
            ...(account.gradualExposure && { gradualExposure: account.gradualExposure }),
        };
    }

    isSystemEntity() {
        return this.getType() === Types.SYSTEM;
    }

    isEngineEntity() {
        return this.getType() === Types.ENGINE;
    }

    isRuntimeEnvironmentEntity() {
        return this.getType() === Types.RUNTIME_ENVIRONMENT;
    }

    isUserEntity() {
        return this.getType() === Types.USER;
    }

    isAgentEntity() {
        return this.getType() === Types.AGENT;
    }

    isExecutionContextEntity() {
        return this.getType() === Types.EXECUTION_CONTEXT;
    }

    // eslint-disable-next-line class-methods-use-this
    isAccountAdmin() {
        return false; // default implementation - should be overridden if needed
    }

    getName() {
        return _.get(this, '_entity.name', this.getType());
    }

    set userAgent(userAgent) {
        const splittedUserAgent = (userAgent || '').split(/-v/);

        if (splittedUserAgent.length === 2) {
            const type = splittedUserAgent[0];
            const version = splittedUserAgent[1];

            if (semver.valid(version) && _.find(USER_AGENT_TYPES, t => t === type.toLowerCase())) {
                this._entity.userAgent = {
                    full: userAgent,
                    type,
                    version
                };
                return;
            }
        }

        this._entity.userAgent = {
            full: userAgent
        };
    }

    hasRecognizedUserAgent() {
        return _.has(this, '_entity.userAgent.type') && _.has(this, '_entity.userAgent.version');
    }

    hasUserAgent() {
        return _.has(this, '_entity.userAgent.full');
    }

    get userAgent() {
        if (!this.hasUserAgent()) {
            throw new Error('Auth entity does not have a user agent');
        }

        return this._entity.userAgent.full;
    }

    getUserAgentType() {
        if (!this.hasUserAgent()) {
            throw new Error('Auth entity does not have a user agent');
        }

        if (!this.hasRecognizedUserAgent()) {
            throw new Error('Auth entity does not have a recognized user agent');
        }

        return this._entity.userAgent.type;
    }

    getUserAgentVersion() {
        if (!this.hasUserAgent()) {
            throw new Error('Auth entity does not have a user agent');
        }

        if (!this.hasRecognizedUserAgent()) {
            throw new Error('Auth entity does not have a recognized user agent');
        }

        return this._entity.userAgent.version;
    }

    getFeatures(accountId, internalOnly = false) { //eslint-disable-line
        return Promise.resolve()
            .then(() => {
                if (this.isSystemEntity() && !accountId) {
                    throw new Error(`System entity must to provide accountId`);
                }
                if (!this.isSystemEntity() && this.getActiveAccountId() === accountId) {
                    accountId = undefined;
                }
                if (!accountId && !internalOnly) {
                    const activeAccount = this.getActiveAccount();
                    if (_.get(activeAccount, 'features')) {
                        return activeAccount.features;
                    }
                }
                const qs = {};
                if (accountId) {
                    qs.accountId = accountId;
                }
                if (internalOnly) {
                    qs.internalOnly = true;
                }
                const request = getInternalRequestLogic();
                return request({
                    serviceConfig: apiService,
                    path: `/api/features-manager/account`,
                    requestOptions: {
                        method: 'GET',
                        qs,
                    },
                })
                    .then(res => res.body);
            });
    }

    evaluateFeatureFlag(feature, accountId, fallback) { //eslint-disable-line
        return Promise.resolve()
            .then(() => {
                if (this.isSystemEntity() && !accountId) {
                    throw new Error(`System entity must to provide accountId`);
                }
                if (!this.isSystemEntity() && this.getActiveAccountId() === accountId) {
                    accountId = undefined;
                }
                if (!accountId) {
                    const activeAccount = this.getActiveAccount();
                    if (_.get(activeAccount, 'features')) {
                        const featureValue = activeAccount.features[feature];
                        if (_.isUndefined(featureValue)) {
                            throw new Error(`Feature ${feature} is not valid`);
                        }
                        return featureValue;
                    }
                }
                const request = getInternalRequestLogic();
                const qs = {};
                if (accountId) {
                    qs.accountId = accountId;
                }
                return request({
                    serviceConfig: apiService,
                    path: `/api/features-manager/account/evaluate/${feature}`,
                    requestOptions: {
                        method: 'GET',
                        qs
                    },
                }).then(res => res.body);
            })
            .catch((err) => {
                logger.error(`Error was occurred during ${feature} flag evaluation:`, err);
                if (!_.isUndefined(fallback)) {
                    logger.warn(`Fallback was provided for feature ${feature} return ${fallback}`);
                    return fallback;
                }
                logger.warn(`Fallback was not provided for feature ${feature} return false`);
                return false;
            });
    }

    getExtendedConfiguration(runtime) {
        return Promise.resolve()
            .then(() => {
                runtime.activeAccount = this.getActiveAccount();
                const request = getInternalRequestLogic();
                return request({
                    serviceConfig: apiService,
                    path: `/api/features-manager/runtime/extended`,
                    requestOptions: {
                        method: 'GET',
                        qs: {
                            runtime
                        },
                    },
                })
                    .then(res => res.body);
            });
    }

    getActiveAccount() {
        throw new Error(`Entity ${this._entity.type} does not have an active account`);
    }

    getActiveAccountId() {
        const account = this.getActiveAccount();
        return _.get(account, '_id') || _.get(account, 'id');
    }

    getFeaturesOfActiveAccount() {
        const activeAccount = this.getActiveAccount();
        return _.get(activeAccount, 'features');
    }

    _getPartialJson() {
        return {
            type: this.getType(),
        };
    }
}


module.exports = Entity;
