const _ = require('lodash');
const Entity = require('./Entity');
const Types = require('./Types');
const { getInternalRequestLogic, paymentsService } = require('./internalRequest');

const requiredParams = [
    'userName',
    '_id',
    'account'
];

/**
 * Should not be used directly with constructor, rather only through factory
 * @param data
 */

/**
 * Basic user entity should have
 * @param name: String
 * @param _id: String
 * @param account: [Object]
 */

class User extends Entity {
    /**
     * Use this static method to create new User entity from given user model
     * @param user
     * @return {User}
     */
    static constructFromUserEntity(user) {
        if (!requiredParams.every(param => _.has(user, param))) {
            throw new Error('Could not construct User entity with given data');
        }

        const userId = user._id.toString();
        const data = {
            type: Types.USER,
            name: user.userName,
            _id: userId,
            accounts: user.account.map(account => this.extractAccountData(account, userId)),
            isSystemAdmin: (user.roles || []).includes('Admin'),
            apiKeyCreatedBySystem: _.get(user, 'apiKeyCreatedBySystem', false),
        };

        return new User(data);
    }

    setActiveAccount(name, id, accountFeatures) {
        const account = this.getAccountById(id);
        if (!account) {
            throw new Error('The user is not part of the provided account');
        }

        if (accountFeatures) {
            account.features = accountFeatures;
        }

        _.set(this, '_entity.activeAccount', account);
    }

    getActiveAccount() {
        return _.get(this, '_entity.activeAccount', {});
    }

    getAccountByName(name) {
        return this._entity.accounts.find(a => a.name === name);
    }

    getAccountById(id) {
        return this._entity.accounts.find(a => a._id === id);
    }

    isAccountAdmin(id) {
        const account = this.getAccountById(id);

        return Boolean(account && account.isAdmin);
    }

    isAccountMember(id) {
        const account = this.getAccountById(id);

        return Boolean(account);
    }

    getTeamsByAccountName(name) {
        const account = this.getAccountByName(name);
        let teams;
        if (account) {
            teams = account.teams;
        }
        return teams;
    }

    getTeamsByAccountId(id) {
        const account = this.getAccountById(id);
        let teams;
        if (account) {
            teams = account.teams;
        }
        return teams;
    }

    getFeaturesByAccountName(name) {
        const account = this.getAccountByName(name);
        return this.getFeatures(account);
    }

    getFeaturesByAccountId(id) {
        const account = this.getAccountById(id);
        return this.getFeatures(account);
    }

    getPlanByAccountId(accountId) { // eslint-disable-line
        const request = getInternalRequestLogic();
        return request({
            serviceConfig: paymentsService,
            path: `/api/${accountId}/plan`,
            requestOptions: {
                method: 'GET'
            }
        })
            .then(res => res.body)
            .catch((err) => {
                if (!err.toString().includes('Service Unavailable')) {
                    throw err;
                }
            });
    }

    getUserId() {
        return this._entity._id;
    }

    getUserName() {
        return this._entity.name;
    }

    isSystemAdmin() {
        return this._entity.isSystemAdmin;
    }

    isApiKeyGeneratedBySystem() {
        return this._entity.apiKeyCreatedBySystem;
    }

    get id() { // alias for consistency with execution-context entity
        return this.getUserId();
    }

    get name() { // alias for consistency with execution-context entity
        return this.getUserName();
    }

    _getPartialJson() {
        return Object.assign(super._getPartialJson(), {
            id: this.getUserId(),
            name: this.getUserName(),
            activeAccount: this.getActiveAccount()
        });
    }
}


module.exports = User;
