'use strict';

/* eslint-disable no-unused-expressions */
/* eslint-disable no-undef */

const _ = require('lodash');
const chai = require('chai');
const sinonChai = require('sinon-chai');
const sinon = require('sinon');

const expect = chai.expect;
const Entity = require('../Entity');
const User = require('../User');
const System = require('../System');
const EntityTypes = require('./../Types');
const USER_AGENT_TYPES = require('./../UserAgentsTypes');

EntityTypes.TEST = 'fake-type-for-testing';

chai.use(sinonChai);

describe('System unit ', () => {

    it('should fail in case type is missing', () => {
        try {
            new Entity({});
        } catch (err) {
            expect(err.toString()).to.contain('Could not create an entity without a type filed');
            return;
        }
        throw new Error('should have failed');
    });

    it('should return true if entity is system', () => {

        const userJson = {
            type: 'user',
            _id: 'fake-user-id-1',
            userName: 'fake-userName',
            account: [
                {
                    name: 'fake-account-name-1',
                    _id: 'fake-accountId-1',
                    admins: ['fake-user-id-2', 'fake-user-id-1']
                },
                {
                    name: 'fake-account-name-2',
                    _id: 'fake-accountId-2',
                    admins: ['fake-user-id-2', 'fake-user-id-3']
                }
            ]
        };
        const user = User.constructFromUserEntity(userJson);
        expect(user.isSystemEntity()).to.equal(false);
    });

    it('should return false if entity is not system', () => {
        const system = new System('microservice');
        expect(system.isSystemEntity()).to.equal(true);
    });

    it('Should get json object', () => {
        const entity = new Entity({ type: EntityTypes.TEST });
        expect(entity.toJson()).to.be.deep.equal({
            type: EntityTypes.TEST,
        });
    });


    it('Should get partial json object', () => {
        const entity = new Entity({ type: EntityTypes.TEST });
        entity._getPartialJson = sinon.spy(entity._getPartialJson);
        entity.toJson({
            partial: true,
        });
        expect(entity._getPartialJson).to.have.been.calledOnce;
    });

    describe('user agent', () => {

        describe('hasRecognizedUserAgent', () => {

            describe('true', () => {

                it('hasRecognizedUserAgent should return true in case user agent was set with a recognized user-agent', () => {
                    const entity = new Entity({ type: EntityTypes.TEST });
                    entity.userAgent = `${USER_AGENT_TYPES.CLI}-v1.0.0`;
                    expect(entity.hasRecognizedUserAgent()).to.equal(true);
                });

                it('hasRecognizedUserAgent should return true in case user agent was set with a recognized user-agent', () => {
                    const entity = new Entity({ type: EntityTypes.TEST });
                    entity.userAgent = `${(USER_AGENT_TYPES.CLI).toUpperCase()}-v1.0.0`;
                    expect(entity.hasRecognizedUserAgent()).to.equal(true);
                });
            });

            describe('false', () => {

                it('hasRecognizedUserAgent should return false in case user agent was set with a non recognized user-agent', () => {
                    const entity = new Entity({ type: EntityTypes.TEST });
                    entity.userAgent = `${USER_AGENT_TYPES.CLI}-v1.0`;
                    expect(entity.hasRecognizedUserAgent()).to.equal(false);
                });

                it('hasRecognizedUserAgent should return false in case user agent was set with a non recognized user-agent', () => {
                    const entity = new Entity({ type: EntityTypes.TEST });
                    entity.userAgent = `unrecognized-v1.0.0`;
                    expect(entity.hasRecognizedUserAgent()).to.equal(false);
                });

                it('hasRecognizedUserAgent should return false in case user agent was set with a non recognized user-agent', () => {
                    const entity = new Entity({ type: EntityTypes.TEST });
                    entity.userAgent = `${USER_AGENT_TYPES.CLI}-1.0.0`;
                    expect(entity.hasRecognizedUserAgent()).to.equal(false);
                });
            });

        });

        describe('hasUserAgent', () => {

            describe('true', () => {

                it('hasUserAgent should return true in case user agent was set and is recognized', () => {
                    const entity = new Entity({ type: EntityTypes.TEST });
                    entity.userAgent = `${USER_AGENT_TYPES.CLI}-v1.0.0`;
                    expect(entity.hasRecognizedUserAgent()).to.equal(true);
                    expect(entity.hasUserAgent()).to.equal(true);
                });

                it('hasUserAgent should return true in case user agent was set and is not recognized', () => {
                    const entity = new Entity({ type: EntityTypes.TEST });
                    entity.userAgent = `${USER_AGENT_TYPES.CLI}-v-1.0`;
                    expect(entity.hasRecognizedUserAgent()).to.equal(false);
                    expect(entity.hasUserAgent()).to.equal(true);
                });

            });

            describe('false', () => {

                it('hasUserAgent should return false in case user agent was not set', () => {
                    const entity = new Entity({ type: EntityTypes.TEST });
                    expect(entity.hasUserAgent()).to.equal(false);
                });
            });

        });

        describe('get user agent', () => {

            describe('positive', () => {

                it('should return user agent if user agent is recognized', () => {
                    const entity = new Entity({ type: EntityTypes.TEST });
                    const userAgent = `${USER_AGENT_TYPES.CLI}-v1.0.0`;
                    entity.userAgent = userAgent;
                    expect(entity.hasRecognizedUserAgent()).to.equal(true);
                    expect(entity.userAgent).to.equal(userAgent);
                });

                it('should return user agent if user agent is not recognized', () => {
                    const entity = new Entity({ type: EntityTypes.TEST });
                    const userAgent = `${USER_AGENT_TYPES.CLI}-v1.0`;
                    entity.userAgent = userAgent;
                    expect(entity.hasRecognizedUserAgent()).to.equal(false);
                    expect(entity.userAgent).to.equal(userAgent);
                });

            });

            describe('negative', () => {

                it('should throw an error in case of trying to get user agent when it does not exist', () => {
                    const entity = new Entity({ type: EntityTypes.TEST });
                    let userAgent;
                    try {
                        userAgent = entity.userAgent;
                    } catch (err) {
                        expect(err.toString()).to.equal('Error: Auth entity does not have a user agent');
                        return;
                    }

                    throw new Error(`should have failed, but userAgent is: ${userAgent}`);
                });

            });

        });

        describe('user agent type', () => {

            describe('positive', () => {

                it('should return user agent type', () => {
                    const entity = new Entity({ type: EntityTypes.TEST });
                    const userAgent = `${USER_AGENT_TYPES.CLI}-v1.0.0`;
                    entity.userAgent = userAgent;
                    expect(entity.hasRecognizedUserAgent()).to.equal(true);
                    expect(entity.getUserAgentType()).to.equal(USER_AGENT_TYPES.CLI);
                });

                it('should return user agent type', () => {
                    const entity = new Entity({ type: EntityTypes.TEST });
                    const userAgent = `${USER_AGENT_TYPES.ENGINE}-v1.0.0`;
                    entity.userAgent = userAgent;
                    expect(entity.hasRecognizedUserAgent()).to.equal(true);
                    expect(entity.getUserAgentType()).to.equal(USER_AGENT_TYPES.ENGINE);
                });

                it('should return user agent type', () => {
                    const entity = new Entity({ type: EntityTypes.TEST });
                    const userAgent = `${USER_AGENT_TYPES.UI}-v1.0.0`;
                    entity.userAgent = userAgent;
                    expect(entity.hasRecognizedUserAgent()).to.equal(true);
                    expect(entity.getUserAgentType()).to.equal(USER_AGENT_TYPES.UI);
                });

                it('should return user agent type', () => {
                    const entity = new Entity({ type: EntityTypes.TEST });
                    const userAgent = `${USER_AGENT_TYPES.JS_SDK}-v1.0.0`;
                    entity.userAgent = userAgent;
                    expect(entity.hasRecognizedUserAgent()).to.equal(true);
                    expect(entity.getUserAgentType()).to.equal(USER_AGENT_TYPES.JS_SDK);
                });

            });

            describe('negative', () => {

                it('should throw an error in case of trying to get user agent type when it does not exist', () => {
                    const entity = new Entity({ type: EntityTypes.TEST });
                    expect(entity.hasUserAgent()).to.equal(false);

                    let userAgentType;
                    try {
                        userAgentType = entity.getUserAgentType();
                    } catch (err) {
                        expect(err.toString()).to.equal('Error: Auth entity does not have a user agent');
                        return;
                    }

                    throw new Error(`should have failed, but userAgent type is: ${userAgentType}`);
                });

                it('should throw an error in case of trying to get user agent type when user agent exists but is not recognized', () => {
                    const entity = new Entity({ type: EntityTypes.TEST });
                    const userAgent = `${USER_AGENT_TYPES.JS_SDK}-v1.0`;
                    entity.userAgent = userAgent;
                    expect(entity.hasUserAgent()).to.equal(true);
                    expect(entity.hasRecognizedUserAgent()).to.equal(false);

                    let userAgentType;
                    try {
                        userAgentType = entity.getUserAgentType();
                    } catch (err) {
                        expect(err.toString()).to.equal('Error: Auth entity does not have a recognized user agent');
                        return;
                    }

                    throw new Error(`should have failed, but userAgent type is: ${userAgentType}`);
                });

            });

        });

        describe('user agent version', () => {

            describe('positive', () => {

                it('should return user agent type', () => {
                    const entity = new Entity({ type: EntityTypes.TEST });
                    const version = '1.0.0';
                    const userAgent = `${USER_AGENT_TYPES.CLI}-v${version}`;
                    entity.userAgent = userAgent;
                    expect(entity.hasRecognizedUserAgent()).to.equal(true);
                    expect(entity.getUserAgentVersion()).to.equal(version);
                });

                it('should return user agent type', () => {
                    const entity = new Entity({ type: EntityTypes.TEST });
                    const version = '1.0.0';
                    const userAgent = `${USER_AGENT_TYPES.ENGINE}-v${version}`;
                    entity.userAgent = userAgent;
                    expect(entity.hasRecognizedUserAgent()).to.equal(true);
                    expect(entity.getUserAgentVersion()).to.equal(version);
                });

                it('should return user agent type', () => {
                    const entity = new Entity({ type: EntityTypes.TEST });
                    const version = '1.0.0';
                    const userAgent = `${USER_AGENT_TYPES.UI}-v${version}`;
                    entity.userAgent = userAgent;
                    expect(entity.hasRecognizedUserAgent()).to.equal(true);
                    expect(entity.getUserAgentVersion()).to.equal(version);
                });

                it('should return user agent type', () => {
                    const entity = new Entity({ type: EntityTypes.TEST });
                    const version = '1.0.0';
                    const userAgent = `${USER_AGENT_TYPES.JS_SDK}-v${version}`;
                    entity.userAgent = userAgent;
                    expect(entity.hasRecognizedUserAgent()).to.equal(true);
                    expect(entity.getUserAgentVersion()).to.equal(version);
                });

                it('should return user agent type', () => {
                    const entity = new Entity({ type: EntityTypes.TEST });
                    const version = '1.1.1';
                    const userAgent = `${USER_AGENT_TYPES.JS_SDK}-v${version}`;
                    entity.userAgent = userAgent;
                    expect(entity.hasRecognizedUserAgent()).to.equal(true);
                    expect(entity.getUserAgentVersion()).to.equal(version);
                });

            });

            describe('negative', () => {


                it('should throw an error in case of trying to get user agent version when it does not exist', () => {
                    const entity = new Entity({ type: EntityTypes.TEST });
                    expect(entity.hasUserAgent()).to.equal(false);

                    let userAgentVersion;
                    try {
                        userAgentVersion = entity.getUserAgentVersion();
                    } catch (err) {
                        expect(err.toString()).to.equal('Error: Auth entity does not have a user agent');
                        return;
                    }

                    throw new Error(`should have failed, but userAgent version is: ${userAgentVersion}`);
                });

                it('should throw an error in case of trying to get user agent version when user agent exists but is not recognized', () => {
                    const entity = new Entity({ type: EntityTypes.TEST });
                    const userAgent = `${USER_AGENT_TYPES.JS_SDK}-v1.0`;
                    entity.userAgent = userAgent;
                    expect(entity.hasUserAgent()).to.equal(true);
                    expect(entity.hasRecognizedUserAgent()).to.equal(false);

                    let userAgentVersion;
                    try {
                        userAgentVersion = entity.getUserAgentVersion();
                    } catch (err) {
                        expect(err.toString()).to.equal('Error: Auth entity does not have a recognized user agent');
                        return;
                    }

                    throw new Error(`should have failed, but userAgent version is: ${userAgentVersion}`);
                });

            });

        });

    });
});

